"""
Sentinel-Toolkit
================

`Sentinel-Toolkit <https://github.com/sentinel-toolkit/sentinel-toolkit>`__ is an open-source
`Python <https://www.python.org/>`__ package providing various utility tools for working
with Sentinel-2 satellite images. For example, you can use sentinel_toolkit.S2Srf class
as a python wrapper of a given S2_SRF Excel file to read the spectral responses into ndarray.
"""

from sentinel_toolkit import ecostress
from sentinel_toolkit import srf
from sentinel_toolkit import colorimetry
from sentinel_toolkit import converter
from sentinel_toolkit import product



"""
s2_metadata provides the class S2ProductMetadata that acts as a python wrapper
of the Sentinel-2 product metadata.
"""

from pathlib import Path
import xml.etree.ElementTree as ET

import numpy as np


class S2ProductMetadata:
    """
    Provides methods for reading Sentinel-2 product metadata.
    """

    _SENSING_DATE_TAG = "DATATAKE_SENSING_START"

    _SPECIAL_VALUES_TAG = "Special_Values"
    _SPECIAL_VALUE_TEXT_TAG = "SPECIAL_VALUE_TEXT"
    _SPECIAL_VALUE_INDEX_TAG = "SPECIAL_VALUE_INDEX"
    _NODATA_TEXT_VALUE = "NODATA"

    _BOA_QUANTIFICATION_VALUE_TAG = "BOA_QUANTIFICATION_VALUE"
    _BOA_OFFSET_VALUES_LIST_TAG = "BOA_ADD_OFFSET_VALUES_LIST"
    _BOA_OFFSET_TAG = "BOA_ADD_OFFSET"
    _BOA_OFFSET_BAND_ID_ATTRIBUTE = "band_id"

    _SOLAR_IRRADIANCE_TAG = "SOLAR_IRRADIANCE"
    _SOLAR_IRRADIANCE_BAND_ID_ATTRIBUTE = "bandId"

    def __init__(self, filename):
        self.filename = filename
        file = Path(filename)
        self.tree = ET.parse(file)

    def get_filename(self):
        """
        Retrieves the metadata filename.

        Returns
        -------
        output : str
            The metadata filename.
        """
        return self.filename

    def get_sensing_date(self):
        """
        Retrieves the sensing date.

        Returns
        -------
        output : str
            The sensing date in format yyyy-mm-dd.
        """
        raw_sensing_date = self.tree.find(f'.//{self._SENSING_DATE_TAG}')
        sensing_date = raw_sensing_date.text.split('T')[0]
        return sensing_date

    def get_nodata_value(self):
        """
        Retrieves the nodata value.

        Returns
        -------
        output : int
            The nodata value, usually 0.
        """
        nodata_element = self.tree.find(
            f'.//{self._SPECIAL_VALUES_TAG}'
            f'[{self._SPECIAL_VALUE_TEXT_TAG} = "{self._NODATA_TEXT_VALUE}"]'
            f'/{self._SPECIAL_VALUE_INDEX_TAG}')
        nodata_value = int(nodata_element.text)
        return nodata_value

    def get_quantification_value(self):
        """
        Retrieves the BOA Quantification Value.

        Returns
        -------
        output : int
            The BOA quantification value, usually 10000.
        """
        boa_quantification_value_element = self.tree.find(
            f'.//{self._BOA_QUANTIFICATION_VALUE_TAG}')
        boa_quantification_value = int(boa_quantification_value_element.text)
        return boa_quantification_value

    def get_band_id_to_offset(self):
        """
        Retrieves a dictionary of band ids to BOA Offsets.

        Returns
        -------
        output : dict
            A dictionary containing the band id to BOA Offset mappings.
        """
        band_id_to_boa_offset = {}

        for boa_add_offset_element in self.tree.iter(self._BOA_OFFSET_TAG):
            band_id = int(boa_add_offset_element.attrib[self._BOA_OFFSET_BAND_ID_ATTRIBUTE])
            band_offset = int(boa_add_offset_element.text)
            band_id_to_boa_offset[band_id] = band_offset

        return band_id_to_boa_offset

    def get_offsets(self, band_ids):
        """
        Retrieves the BOA Offsets for the given band ids.

        Parameters
        ----------
        band_ids : list of int
            the band ids.

        Returns
        -------
        output : ndarray
            An array of the bands' BOA offsets.

        """
        return np.array([self.get_band_id_to_offset()[band_id] for band_id in band_ids])

    def get_band_id_to_solar_irradiance(self):
        """
        Retrieves a dictionary of band ids to solar irradiances.

        Returns
        -------
        output : dict
            A dictionary containing the band id to solar irradiance mappings.
        """
        band_id_to_solar_irradiance = {}

        for solar_irradiance_element in self.tree.iter(self._SOLAR_IRRADIANCE_TAG):
            band_id = int(solar_irradiance_element.attrib[self._SOLAR_IRRADIANCE_BAND_ID_ATTRIBUTE])
            solar_irradiance_value = float(solar_irradiance_element.text)
            band_id_to_solar_irradiance[band_id] = solar_irradiance_value

        return band_id_to_solar_irradiance

    def get_solar_irradiances(self, band_ids):
        """
        Retrieves the solar irradiances for the given band ids.

        Parameters
        ----------
        band_ids : list of int
            the band ids.

        Returns
        -------
        output : ndarray
            An array of the bands' solar irradiances.
        """
        return np.array([self.get_band_id_to_solar_irradiance()[band_id] for band_id in band_ids])

    def get_normalized_solar_irradiances(self, band_ids):
        """
        Retrieves the normalized solar irradiances for the given band ids.

        Parameters
        ----------
        band_ids : list of int
            the band ids.

        Returns
        -------
        output : ndarray
            An array of the normalized bands' solar irradiances.
        """
        solar_irradiances = self.get_solar_irradiances(band_ids)
        solar_irradiances = solar_irradiances / np.max(solar_irradiances)
        return solar_irradiances



"""
Srf
================

Srf module provides tools for working with the Sentinel-2 SRF.

S2Srf class acts as a python wrapper around a given S2_SRF Excel file.
It provides methods for getting the wavelengths, band data, loading the
whole data into a colour.MultiSpectralDistributions object, etc.

S2SrfOptions is a wrapper around the satellite, band names and wavelength
range properties.
"""

from sentinel_toolkit.srf.s2_srf import S2Srf
from sentinel_toolkit.srf.s2_srf import S2SrfOptions



"""
Ecostress
================

Ecostress module provides tools for working with NASA Ecostress library.

It utilizes the library `spectral` and provides a script for loading the
Ecostress spectral data directory into an SQLite database.

It also provides the class Ecostress that adds some convenient methods
for working with spectral.EcostressDatabase
"""

from sentinel_toolkit.ecostress.ecostress_db_generator import generate_ecostress_db
from sentinel_toolkit.ecostress.ecostress import Ecostress



"""
Colorimetry
================

Colorimetry module provides tools for converting dn values or
spectral distribution to sentinel.
"""

from sentinel_toolkit.colorimetry.sentinel_values import SpectralData

from sentinel_toolkit.colorimetry.sentinel_values import sd_to_sentinel_colour
from sentinel_toolkit.colorimetry.sentinel_values import sd_to_sentinel_direct_colour

from sentinel_toolkit.colorimetry.sentinel_values import sd_to_sentinel_numpy
from sentinel_toolkit.colorimetry.sentinel_values import sd_to_sentinel_direct_numpy

from sentinel_toolkit.colorimetry.sentinel_values import dn_to_sentinel

from sentinel_toolkit.colorimetry.illuminants import D65_360_830_1NM_DISTRIBUTION
from sentinel_toolkit.colorimetry.illuminants import D65_360_830_1NM_VALUES



"""
sentinel_values provides methods for converting
a spectral distribution to sentinel responses.
"""

from collections import namedtuple

import colour
import numpy as np

from sentinel_toolkit.colorimetry.illuminants import D65_360_830_1NM_DISTRIBUTION
from sentinel_toolkit.colorimetry.illuminants.d65 import D65_360_830_1NM_VALUES

SpectralData = namedtuple("SpectralData", "wavelengths spectral_responses")


def sd_to_sentinel_colour(spectral_distribution,
                          s2_srf,
                          s2_srf_options,
                          illuminant=None):
    """
    Converts a spectral distribution to Sentinel-2 responses.

    Parameters
    ----------
    spectral_distribution : colour.SpectralDistribution
        The spectral distribution to convert.
    s2_srf : sentinel_toolkit.S2Srf
        The Sentinel-2 spectral response functions.
    s2_srf_options : S2SrfOptions
        The satellite, band names and wavelength range.
        If satellite is missing, satellite 'A' will be used.
        If band names are missing, all band names will be used.
        If wavelength range is missing, (360, 830) will be used.
    illuminant : colour.SpectralDistribution
        The illuminant to apply.
        If missing, default to D65 360-830 nm.

    Returns
    -------
    output : ndarray
        The Sentinel-2 spectral responses.
    """
    bands_responses = s2_srf.get_bands_responses(s2_srf_options)
    return sd_to_sentinel_direct_colour(spectral_distribution, bands_responses, illuminant)


def sd_to_sentinel_direct_colour(spectral_distribution, bands_responses, illuminant=None):
    """
    Converts a spectral distribution to Sentinel-2 responses.

    Parameters
    ----------
    spectral_distribution : colour.SpectralDistribution
        The spectral distribution ro convert.
    bands_responses : ndarray
        The bands_responses functions as a 2D ndarray.
    illuminant : colour.SpectralDistribution
        The illuminant to apply.
        If missing, default to D65 360-830 nm.

    Returns
    -------
    output : ndarray
        The Sentinel-2 spectral responses.
    """
    if illuminant is None:
        shape = spectral_distribution.shape
        illuminant = colour.SpectralDistribution(D65_360_830_1NM_DISTRIBUTION).trim(shape)

    row_sum = np.sum(bands_responses, axis=1)
    # Hack for solving division by zero optimally
    row_sum[row_sum == 0] = 1
    bands_srf = bands_responses / row_sum[:, None]

    sd_i = spectral_distribution.values * illuminant.values

    return np.dot(bands_srf, sd_i)


def sd_to_sentinel_numpy(spectral_data,
                         s2_srf,
                         s2_srf_options,
                         illuminant=None):
    """
    Converts a spectral distribution to Sentinel-2 responses.

    Parameters
    ----------
    spectral_data : SpectralData (tuple) of ndarray
        The spectral distribution.
    s2_srf : sentinel_toolkit.S2Srf
        The Sentinel-2 spectral response functions.
    s2_srf_options : S2SrfOptions
        The satellite, band names and wavelength range.
        If satellite is missing, satellite 'A' will be used.
        If band ids are missing, all band ids will be used.
        If wavelength range is missing, (360, 830) will be used.
    illuminant : colour.SpectralDistribution
        The illuminant to apply.
        If missing, default to D65 360-830 nm.

    Returns
    -------
    output : ndarray
        The Sentinel-2 spectral responses.
    """
    bands_responses = s2_srf.get_bands_responses(s2_srf_options)
    return sd_to_sentinel_direct_numpy(spectral_data, bands_responses, illuminant)


def sd_to_sentinel_direct_numpy(spectral_data, bands_responses, illuminant=None):
    """
    Converts a spectral distribution to Sentinel-2 responses.

    Note that currently there is no reshaping, so all the arrays
    should have valid dimensions.

    Parameters
    ----------
    spectral_data : SpectralData (tuple) of ndarray
        The wavelengths and spectral_responses.
    bands_responses : ndarray
        The bands_responses functions as a 2D ndarray.
    illuminant : ndarray
        The illuminant to apply.
        If missing, default to D65 360-830 nm.

    Returns
    -------
    output : ndarray
        The Sentinel-2 spectral responses.
    """
    min_wavelength = int(spectral_data.wavelengths[0])
    max_wavelength = int(spectral_data.wavelengths[-1])

    if illuminant is None:
        illuminant = D65_360_830_1NM_VALUES[min_wavelength - 360: max_wavelength - 359]

    row_sum = np.sum(bands_responses, axis=1)
    # Hack for solving division by zero optimally
    row_sum[row_sum == 0] = 1
    bands_srf = bands_responses / row_sum[:, None]

    sd_i = spectral_data.spectral_responses * illuminant

    return np.dot(bands_srf, sd_i)


def dn_to_sentinel(raw_band_data, nodata_value, boa_offset, quantification_value, solar_irradiance):
    """
    Converts Sentinel-2 DN values to sentinel responses.

    Parameters
    ----------
    nodata_value : int
        The nodata pixel value.
    raw_band_data : ndarray
        The raw sentinel-2 band(s) pixel values.
    boa_offset : int scalar or ndarray
        The boa offset.
    solar_irradiance : float scalar or ndarray
        The solar irradiance.
    quantification_value : int
        The quantification value.

    Returns
    -------
    output : ndarray
        The band data as sentinel response values.

    """
    raw_band_data_copy = np.copy(raw_band_data)
    nodata_indices = raw_band_data_copy == nodata_value
    raw_band_data_copy += boa_offset
    raw_band_data_copy[nodata_indices] = nodata_value

    sentinel_responses = raw_band_data_copy / quantification_value
    sentinel_responses = sentinel_responses * solar_irradiance
    sentinel_responses = np.clip(sentinel_responses, 0, 1, out=sentinel_responses)

    return sentinel_responses * 100.0



import unittest
from unittest.mock import patch

from sentinel_toolkit.ecostress import generate_ecostress_db


class TestEcostressDbGenerator(unittest.TestCase):
    _DB_FILENAME = "ecostress.db"
    _ECOSTRESS_DATA_DIRECTORY = "ecospeclib-all"

    @patch('spectral.EcostressDatabase.create')
    def test_generate_ecostress_db(self, mock_ecostress_database_create):
        generate_ecostress_db(self._ECOSTRESS_DATA_DIRECTORY, self._DB_FILENAME)
        mock_ecostress_database_create.assert_called_once_with(self._DB_FILENAME, self._ECOSTRESS_DATA_DIRECTORY)


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest import mock
from unittest.mock import patch, mock_open

import numpy as np
from colour import SpectralDistribution

from sentinel_toolkit.colorimetry import SpectralData
from sentinel_toolkit.converter import EcostressToSentinelConverter


class TestConverter(unittest.TestCase):
    _DB_FILENAME = "ecostress.db"
    _ECOSTRESS_DATA_DIRECTORY = "ecospeclib-all"

    _SPECTRAL_DISTRIBUTION = SpectralDistribution(
        {
            438: 0.104,
            439: 0.1042,
            440: 0.1043,
            441: 0.1044,
            442: 0.1045,
            443: 0.1046
        }
    )

    _BANDS_RESPONSES = np.array([
        [0.810893815261278, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0.824198756004815, 0.010315428586995, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0.854158107215052, 0.0300419331664615, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0.870790876711332, 0.0268758905675229, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0.887310969404313, 0.0241431503861076, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0.926199242291321, 0.0202100642531871, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    ]).T

    _EXPECTED_SENTINEL_RESPONSE = [10.985433231270072, 11.086160373966965, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

    _BAND_NAMES = ["S2A_SR_AV_B1",
                   "S2A_SR_AV_B2",
                   "S2A_SR_AV_B3",
                   "S2A_SR_AV_B4",
                   "S2A_SR_AV_B5",
                   "S2A_SR_AV_B6",
                   "S2A_SR_AV_B7",
                   "S2A_SR_AV_B8",
                   "S2A_SR_AV_B8A",
                   "S2A_SR_AV_B9",
                   "S2A_SR_AV_B10",
                   "S2A_SR_AV_B11",
                   "S2A_SR_AV_B12"]

    @patch('sentinel_toolkit.srf.S2Srf')
    @patch('sentinel_toolkit.ecostress.Ecostress')
    @patch('builtins.open', new_callable=mock_open())
    def test_convert_ecostress_to_sentinel(self, mock_open_file, mock_ecostress, mock_s2_srf):
        mock_s2_srf.get_band_names.return_value = self._BAND_NAMES
        mock_s2_srf.get_bands_responses.return_value = self._BANDS_RESPONSES

        mock_ecostress.get_spectrum_ids.return_value = np.array([1])
        spectral_data = SpectralData(self._SPECTRAL_DISTRIBUTION.wavelengths,
                                     self._SPECTRAL_DISTRIBUTION.values)
        mock_ecostress.get_spectral_distribution_numpy.return_value = spectral_data

        converter = EcostressToSentinelConverter(mock_ecostress, mock_s2_srf)
        converter.convert_ecostress_to_sentinel_csv()
        mock_open_file.assert_called_once_with('sentinel_A.csv', 'w', encoding='utf-8')

        band_names_line = ','.join(self._BAND_NAMES)
        csv_heading = f"SpectrumID,{band_names_line}\n"
        csv_sentinel_responses_line = \
            "1,10.985433231270072,11.086160373966965,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0\n"

        mock_open_file.return_value.__enter__().write.assert_has_calls([
            mock.call(csv_heading),
            mock.call(csv_sentinel_responses_line)
        ])


if __name__ == '__main__':
    unittest.main()



"""
ecostress_db_generator provides a way to generate SQLite database
containing the Ecostress spectral library.
ecostress_db_generator.py can be used as a script to generate
the database in the following manner:

python ecostress_db_generator.py -d <ecostress_directory> -o <output_filename>
"""

from argparse import ArgumentParser
from pathlib import Path
from spectral import EcostressDatabase

_DB_FILENAME = "ecostress.db"
_ECOSTRESS_DATA_DIRECTORY = "ecospeclib-all"


def generate_ecostress_db(ecostress_dir=_ECOSTRESS_DATA_DIRECTORY, output_db_filename=_DB_FILENAME):
    """
    Generates SQLite database for a given Ecostress spectral library directory.

    Parameters
    ----------
    ecostress_dir : str
        The Ecostress spectral library directory.
        If missing, default to "ecospeclib-all".
    output_db_filename : str
        The name of the output SQLite database file.
        If missing, default to "ecostress.db".
    """
    EcostressDatabase.create(output_db_filename, ecostress_dir)


def _main():
    args = _parse_args()
    ecostress_dir = args.dir
    output_db_filename = args.out

    ecostress_dir_path = Path(ecostress_dir)
    if not ecostress_dir_path.is_dir():
        error_msg = f'The provided ecostress data directory "{ecostress_dir}" does not exist!'
        raise RuntimeError(error_msg)

    generate_ecostress_db(ecostress_dir, output_db_filename)


def _parse_args():
    parser = ArgumentParser(description="Ecostress SQLite Database Generator")
    parser.add_argument('-d', '--dir', required=False, type=str, default=_ECOSTRESS_DATA_DIRECTORY,
                        help="Ecostress data directory")
    parser.add_argument('-o', '--out', required=False, type=str, default=_DB_FILENAME,
                        help="The filename that will be used to store the SQLite database")

    return parser.parse_args()


if __name__ == "__main__":
    _main()



"""
Illuminants module provides various illuminant spectral data.
"""

from sentinel_toolkit.colorimetry.illuminants.d65 import D65_360_830_1NM_DISTRIBUTION
from sentinel_toolkit.colorimetry.illuminants.d65 import D65_360_830_1NM_VALUES



import os
import unittest

import numpy as np
from colour import MultiSpectralDistributions
from numpy.testing import assert_array_equal

from sentinel_toolkit.srf import S2Srf
from sentinel_toolkit.srf import S2SrfOptions


class TestS2Srf(unittest.TestCase):
    _SRF_FILENAME = os.path.dirname(__file__) + "/test_data/s2a_srf.xlsx"

    _B1_NAME = "S2A_SR_AV_B1"
    _B1_INDEX = 0

    _B2_NAME = "S2A_SR_AV_B2"
    _B2_INDEX = 1

    _EXPECTED_BANDS_RESPONSES_DISTRIBUTION = MultiSpectralDistributions(
        {
            438: (0.810893815261278, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
            439: (0.824198756004815, 0.010315428586995, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
            440: (0.854158107215052, 0.0300419331664615, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
            441: (0.870790876711332, 0.0268758905675229, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
            442: (0.887310969404313, 0.0241431503861076, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
            443: (0.926199242291321, 0.0202100642531871, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
        }
    )

    _WAVELENGTH_RANGE = (438, 439)

    _EXPECTED_ALL_BAND_NAMES = ["S2{}_SR_AV_B1",
                                "S2{}_SR_AV_B2",
                                "S2{}_SR_AV_B3",
                                "S2{}_SR_AV_B4",
                                "S2{}_SR_AV_B5",
                                "S2{}_SR_AV_B6",
                                "S2{}_SR_AV_B7",
                                "S2{}_SR_AV_B8",
                                "S2{}_SR_AV_B8A",
                                "S2{}_SR_AV_B9",
                                "S2{}_SR_AV_B10",
                                "S2{}_SR_AV_B11",
                                "S2{}_SR_AV_B12"]

    def setUp(self):
        self.s2_srf = S2Srf(self._SRF_FILENAME)

    def test_get_wavelengths(self):
        expected = self._EXPECTED_BANDS_RESPONSES_DISTRIBUTION.wavelengths
        actual = self.s2_srf.get_wavelengths()
        assert_array_equal(expected, actual)

    def test_get_all_bands_responses(self):
        expected = self._EXPECTED_BANDS_RESPONSES_DISTRIBUTION.values.T
        actual = self.s2_srf.get_bands_responses()
        assert_array_equal(expected, actual)

    def test_get_bands_responses_satellite(self):
        expected = self._EXPECTED_BANDS_RESPONSES_DISTRIBUTION.values.T
        actual = self.s2_srf.get_bands_responses(S2SrfOptions(satellite='B'))
        assert_array_equal(expected, actual)

    def test_get_specific_bands_responses_with_bands(self):
        b1_values = self._EXPECTED_BANDS_RESPONSES_DISTRIBUTION.values[:, self._B1_INDEX]
        b2_values = self._EXPECTED_BANDS_RESPONSES_DISTRIBUTION.values[:, self._B2_INDEX]

        expected = np.array([b1_values, b2_values])
        actual = self.s2_srf.get_bands_responses(S2SrfOptions(band_ids=[self._B1_INDEX, self._B2_INDEX]))
        assert_array_equal(expected, actual)

    def test_get_bands_responses_with_wavelength_range(self):
        wavelengths = self._EXPECTED_BANDS_RESPONSES_DISTRIBUTION.wavelengths
        mask = (wavelengths >= self._WAVELENGTH_RANGE[0]) & (wavelengths <= self._WAVELENGTH_RANGE[1])
        expected = self._EXPECTED_BANDS_RESPONSES_DISTRIBUTION.values[mask].T

        actual = self.s2_srf.get_bands_responses(S2SrfOptions(wavelength_range=self._WAVELENGTH_RANGE))
        assert_array_equal(expected, actual)

    def test_get_all_bands_names_satellite_A(self):
        expected_band_names = list(map(lambda b: b.format('A'), self._EXPECTED_ALL_BAND_NAMES))
        actual = self.s2_srf.get_band_names()
        assert_array_equal(expected_band_names, actual)

    def test_get_bands_names_satellite_A(self):
        band_ids = [1, 2, 3]
        expected_band_names = np.array(list(map(lambda b: b.format('A'), self._EXPECTED_ALL_BAND_NAMES)))[band_ids]
        actual = self.s2_srf.get_band_names(band_ids)
        assert_array_equal(expected_band_names, actual)

    def test_get_all_bands_names_satellite_B(self):
        expected_band_names = list(map(lambda b: b.format('B'), self._EXPECTED_ALL_BAND_NAMES))
        actual = self.s2_srf.get_band_names(satellite='B')
        assert_array_equal(expected_band_names, actual)

    def test_get_bands_names_satellite_B(self):
        band_ids = [1, 2, 3]
        expected_band_names = np.array(list(map(lambda b: b.format('B'), self._EXPECTED_ALL_BAND_NAMES)))[band_ids]
        actual = self.s2_srf.get_band_names(band_ids, satellite='B')
        assert_array_equal(expected_band_names, actual)

    def test_get_bands_responses_distribution_satellite_A(self):
        actual = self.s2_srf.get_bands_responses_distribution()
        self.assertEqual(self._EXPECTED_BANDS_RESPONSES_DISTRIBUTION, actual)

    def test_get_bands_responses_distribution_satellite_B(self):
        actual = self.s2_srf.get_bands_responses_distribution(S2SrfOptions(satellite='B'))
        self.assertEqual(self._EXPECTED_BANDS_RESPONSES_DISTRIBUTION, actual)


if __name__ == '__main__':
    unittest.main()



"""
s2_product provides the class S2Product that provides various
tools for working with Sentinel-2 product data.
"""

import re
from pathlib import Path

import numpy as np
import rasterio

from rasterio.enums import Resampling

from sentinel_toolkit.colorimetry import dn_to_sentinel
from . import product_metadata


class S2Product:
    """
    Provides methods for working with a Sentinel-2 product.
    """

    _METADATA_FILENAME_REGEX = 'MTD_MSIL*.xml'
    _BAND_ID_TO_NAME = {
        0: 'B01',
        1: 'B02',
        2: 'B03',
        3: 'B04',
        4: 'B05',
        5: 'B06',
        6: 'B07',
        7: 'B08',
        8: 'B8A',
        9: 'B09',
        10: 'B10',
        11: 'B11',
        12: 'B12',
    }
    _R10M_FILE_REGEX = ".*10m.jp2"
    _R20M_FILE_REGEX = ".*20m.jp2"
    _R60M_FILE_REGEX = ".*60m.jp2"
    _JP2_DRIVER_NAME = "JP2OpenJPEG"

    def __init__(self, product_directory_name):
        self.directory_name = product_directory_name
        metadata_filename = self._get_metadata_filename()
        self.metadata = product_metadata.S2ProductMetadata(metadata_filename)

    def _get_metadata_filename(self):
        metadata_filenames = self._find_filenames(
            self._METADATA_FILENAME_REGEX,
            recursive=False
        )

        metadata_filenames_count = len(metadata_filenames)
        if metadata_filenames_count == 1:
            return metadata_filenames[0]

        raise RuntimeError(f'Number of found metadata files matching '
                           f'regex {self._METADATA_FILENAME_REGEX} '
                           f'is {metadata_filenames_count}.')

    def _find_filenames(self, regex, recursive=True):
        if recursive:
            file_paths = Path(self.directory_name).rglob(regex)
        else:
            file_paths = Path(self.directory_name).glob(regex)

        filenames = []
        for file_path in file_paths:
            filenames.append(str(file_path))

        return filenames

    def get_directory_name(self):
        """
        Retrieves the product directory name.

        Returns
        -------
        output : str
            The product directory name.
        """
        return self.directory_name

    def get_metadata(self):
        """
        Retrieves the metadata of the product.

        Returns
        -------
        output : sentinel_toolkit.product.metadata.metadata.S2ProductMetadata
            The metadata of the product.
        """
        return self.metadata

    def get_dn_source(self, band_id):
        """
        Retrieves the dn source.

        When multiple files with different Spatial Resolution
        are found, the best resolution is selected.

        Parameters
        ----------
        band_id : int
            The band id.

        Returns
        -------
        output : rasterio.io.DatasetReader
            A rasterio.io.DatasetReader object.
        """
        band_name = self._BAND_ID_TO_NAME[band_id]
        if band_name is None:
            raise RuntimeError(f"Unsupported band id {band_id}.")

        bands_filenames = self._find_filenames(f"*{band_name}*.jp2",
                                               recursive=True)
        if len(bands_filenames) == 0:
            raise RuntimeError(f"Number of found files for band id {band_id} "
                               f"with band name {band_name} is 0.")

        if len(bands_filenames) == 1:
            band_filename = bands_filenames[0]
        else:
            band_filename = self._get_best_band_filename(bands_filenames)
            if band_filename is None:
                raise RuntimeError(f"Cannot find a band file matching regex "
                                   f"{self._R10M_FILE_REGEX} or "
                                   f"{self._R20M_FILE_REGEX} or "
                                   f"{self._R60M_FILE_REGEX}.")

        band_source = rasterio.open(band_filename,
                                    driver=self._JP2_DRIVER_NAME)
        return band_source

    def _get_best_band_filename(self, bands_filenames):
        for band_filename in bands_filenames:
            if re.match(self._R10M_FILE_REGEX, band_filename):
                return band_filename

        for band_filename in bands_filenames:
            if re.match(self._R20M_FILE_REGEX, band_filename):
                return band_filename

        for band_filename in bands_filenames:
            if re.match(self._R60M_FILE_REGEX, band_filename):
                return band_filename

        return None

    def get_band_id_to_dn_source(self, band_ids):
        """
        Retrieves a dictionary of band ids to DN sources.

        Parameters
        ----------
        band_ids : list of int
            The band ids.

        Returns
        -------
        output : dict
            A dictionary containing the band id to DN source mappings.
        """
        band_id_to_dn = {}
        for band_id in band_ids:
            band_id_to_dn[band_id] = self.get_dn_source(band_id)
        return band_id_to_dn

    def dn_to_sentinel(self, band_ids, out_shape=(10980, 10980)):
        """
        Converts the product dn values to sentinel responses.

        Parameters
        ----------
        band_ids : list of int
            The band ids.

        out_shape : tuple of int
            The out shape.

        Returns
        -------
        output : tuple of rasterio.profiles.Profile and numpy.ndarray
            A tuple containing the modified rasterio profile (with the
            new out_shape and driver GTiff) and the sentinel responses.
        """
        band_id_to_dn_source = self.get_band_id_to_dn_source(band_ids)
        band_id_to_dn = {}

        for band_id, dn_source in band_id_to_dn_source.items():
            band_dn = dn_source.read(1, out_shape=out_shape, resampling=Resampling.cubic)
            band_id_to_dn[band_id] = band_dn

        bands_dn = [band_id_to_dn[band_id] for band_id in band_ids]
        raw_bands_data = np.stack(bands_dn, axis=-1).astype(np.float64)

        metadata = self.metadata
        sentinel_image = dn_to_sentinel(
            raw_bands_data,
            metadata.get_nodata_value(),
            metadata.get_offsets(band_ids),
            metadata.get_quantification_value(),
            metadata.get_normalized_solar_irradiances(band_ids)
        )

        profile = band_id_to_dn_source[band_ids[0]].profile
        self._update_profile(sentinel_image, profile)

        return profile, sentinel_image

    @staticmethod
    def _update_profile(image, profile):
        height, width, count = image.shape

        profile.update({
            'width': width,
            'height': height,
            'count': count,
            'dtype': image.dtype
        })



"""
converter provides the class EcostressToSentinelConverter
that can be used for converting the Ecostress spectral library
to Sentinel-2 responses.
"""

import itertools as it
from argparse import ArgumentParser
from pathlib import Path
from spectral import EcostressDatabase

from sentinel_toolkit.ecostress import Ecostress
from sentinel_toolkit.colorimetry import sd_to_sentinel_numpy
from sentinel_toolkit.srf import S2Srf, S2SrfOptions

_ECOSTRESS_DB_FILENAME = "ecostress.db"
_S2_SRF_FILENAME = "S2-SRF_COPE-GSEG-EOPG-TN-15-0007_3.1.xlsx"


class EcostressToSentinelConverter:
    """
    EcostressToSentinelConverter provides a way to convert
    all the examples from the Ecostress spectral library,
    which contain values in a given wavelength range,
    to Sentinel-2 responses saved to a CSV file.
    """

    def __init__(self, ecostress, s2_srf):
        self.ecostress = ecostress
        self.s2rf = s2_srf

    def convert_ecostress_to_sentinel_csv(self, s2_srf_options=None, illuminant=None, out=None):
        """
        Converts the ecostress library into Sentinel-2 responses csv.

        The converted values are written to a CSV file
        named sentinel_<A or B>.csv.

        Parameters
        ----------
        s2_srf_options : S2SrfOptions
            The satellite, band names and wavelength range.
            If satellite is missing, satellite 'A' will be used.
            If band ids are missing, all band ids will be used.
            If wavelength range is missing, (360, 830) will be used.
        illuminant : ndarray
            The illuminant values.
            If missing, D65 360-830 nm values will be used.
        out : str
            The output filename.
            If missing, "sentinel_<A or B>.csv" will be used.
        """
        if s2_srf_options is None:
            s2_srf_options = S2SrfOptions(satellite='A', wavelength_range=(360, 830))

        satellite, band_ids, wavelength_range = s2_srf_options.unpack()

        band_names = self.s2rf.get_band_names(band_ids)

        output_filename = out if out is not None else f"sentinel_{satellite}.csv"

        spectrum_ids = self.ecostress.get_spectrum_ids(wavelength_range)

        with open(output_filename, 'w', encoding='utf-8') as sentinel_file:
            _write_heading_line(sentinel_file, band_names)
            for spectrum_id in spectrum_ids:
                spectral_data = self.ecostress.get_spectral_distribution_numpy(spectrum_id,
                                                                               wavelength_range)
                wavelength_range = _get_wavelength_range(spectral_data.wavelengths,
                                                         wavelength_range)
                s2_srf_options.wavelength_range = wavelength_range

                sentinel_responses = sd_to_sentinel_numpy(spectral_data,
                                                          self.s2rf,
                                                          s2_srf_options,
                                                          illuminant)
                _write_sentinel_responses_line(sentinel_file, spectrum_id, sentinel_responses)


def _write_heading_line(sentinel_file, band_names):
    band_names_line = ','.join(band_names)
    sentinel_file.write(f"SpectrumID,{band_names_line}\n")


def _get_wavelength_range(spectral_data_wavelengths, wavelength_range):
    spectral_data_min_wavelength = spectral_data_wavelengths[0]
    spectral_data_max_wavelength = spectral_data_wavelengths[-1]

    start = max(wavelength_range[0], spectral_data_min_wavelength)
    end = min(wavelength_range[1], spectral_data_max_wavelength)

    return start, end


def _write_sentinel_responses_line(sentinel_file, spectrum_id, sentinel_responses):
    line = ','.join(it.repeat('{}', len(sentinel_responses) + 1)) + '\n'
    line = line.format(spectrum_id, *sentinel_responses)
    sentinel_file.write(line)


def _main():
    args = _parse_args()
    ecostress_db_filename = args.ecostress_db_filename
    s2_srf_filename = args.s2_srf_filename

    ecostress_db_path = Path(ecostress_db_filename)
    if not ecostress_db_path.is_file():
        error_msg = f'The provided ecostress db filename "{ecostress_db_path}" does not exist!'
        raise RuntimeError(error_msg)

    s2_srf_path = Path(s2_srf_filename)
    if not s2_srf_path.is_file():
        error_msg = f'The provided s2a_srf Excel filename "{s2_srf_path}" does not exist!'
        raise RuntimeError(error_msg)

    ecostress_db = EcostressDatabase(ecostress_db_filename)
    converter = EcostressToSentinelConverter(Ecostress(ecostress_db), S2Srf(s2_srf_filename))

    satellite = args.satellite
    band_ids = args.bands
    wavelength_range = (args.wavelength_start, args.wavelength_end)

    s2_srf_options = S2SrfOptions(satellite=satellite,
                                  band_ids=band_ids,
                                  wavelength_range=wavelength_range)

    out = args.out
    converter.convert_ecostress_to_sentinel_csv(s2_srf_options, out=out)


def _parse_args():
    parser = ArgumentParser(description="Ecostress to Sentinel-2 Converter")
    parser.add_argument('-e',
                        '--ecostress_db_filename',
                        required=False,
                        type=str,
                        default=_ECOSTRESS_DB_FILENAME,
                        help="Ecostress SQLite database filename")
    parser.add_argument('-s2',
                        '--s2_srf_filename',
                        required=False,
                        type=str,
                        default=_S2_SRF_FILENAME,
                        help="Sentinel-2 Spectral Response Functions Excel filename")
    parser.add_argument('-s',
                        '--satellite',
                        required=False,
                        type=str,
                        default='A',
                        help="Sentinel-2 Satellite Identifier - A or B."
                             " By default both satellites will be used.")
    parser.add_argument('-b',
                        '--bands',
                        nargs='*',
                        required=False,
                        type=int,
                        default=[1, 2, 3],
                        help="Sentinel-2 Bands Identifiers List - 0 - 12."
                             " By default bands 1, 2, 3 will be used.")
    parser.add_argument('-ws',
                        '--wavelength_start',
                        required=False,
                        type=int,
                        default=360,
                        help="The wavelength range start. Default is 360.")
    parser.add_argument('-we',
                        '--wavelength_end',
                        required=False,
                        type=int,
                        default=830,
                        help="The wavelength range end. Default is 830.")
    parser.add_argument('-o',
                        '--out',
                        required=False,
                        type=int,
                        default=830,
                        help="The filename that will be used to store the resulting csv.")
    return parser.parse_args()


if __name__ == "__main__":
    _main()



"""
Converter
================

Converter module provides the class EcostressToSentinelConverter
for converting all the examples from the Ecostress spectral library
to Sentinel-2 Responses and writing them to a CSV file.
"""

from sentinel_toolkit.converter.converter import EcostressToSentinelConverter



"""
s2_srf provides the class S2SrfOptions that acts as a python wrapper
of the Sentinel-2 Spectral Response Functions Excel file.
Also, S2SrfOptions dataclass is provided that can be used
by outside callers to configure the desired options.
"""

import warnings

from dataclasses import dataclass
from typing import List, Tuple

import numpy as np
import pandas as pd

from colour import MultiSpectralDistributions

warnings.filterwarnings('ignore', category=UserWarning, module='openpyxl')


@dataclass
class S2SrfOptions:
    """
    Keeps the options that can be passed to some of the S2Srf methods:
    (satellite, band_ids, wavelength_range)
    """
    satellite: str = 'A'
    band_ids: List[int] = None
    wavelength_range: Tuple[int, int] = (360, 830)

    def unpack(self):
        """
        Unpacks the dataclass into satellite, band_names and wavelength_range.

        Returns
        -------
        satellite : str
        band_ids : list of int
        wavelength_range : tuple of int
        """
        return self.satellite, self.band_ids, self.wavelength_range


class S2Srf:
    """
    Provides methods for working with Sentinel-2 SRF Excel file.
    """

    _WAVELENGTH_NAME = "SR_WL"
    _BAND_NAMES = ["S2{}_SR_AV_B1",
                   "S2{}_SR_AV_B2",
                   "S2{}_SR_AV_B3",
                   "S2{}_SR_AV_B4",
                   "S2{}_SR_AV_B5",
                   "S2{}_SR_AV_B6",
                   "S2{}_SR_AV_B7",
                   "S2{}_SR_AV_B8",
                   "S2{}_SR_AV_B8A",
                   "S2{}_SR_AV_B9",
                   "S2{}_SR_AV_B10",
                   "S2{}_SR_AV_B11",
                   "S2{}_SR_AV_B12"]
    _SHEET_NAME = "Spectral Responses (S2{})"

    def __init__(self, filename):
        self.all_band_names = {
            'A': np.array(list(map(lambda b: b.format('A'), self._BAND_NAMES))),
            'B': np.array(list(map(lambda b: b.format('B'), self._BAND_NAMES)))
        }

        self.s2_srf_data = {
            'A': pd.read_excel(filename, sheet_name=self._SHEET_NAME.format('A')),
            'B': pd.read_excel(filename, sheet_name=self._SHEET_NAME.format('B'))
        }

    def get_wavelengths(self, satellite='A'):
        """
        Retrieves the wavelengths.

        Parameters
        ----------
        satellite : str
           The satellite - 'A' or 'B'. If missing, default to 'A'.

        Returns
        -------
        output : ndarray
            An array containing the wavelengths.
        """
        return self.s2_srf_data[satellite][self._WAVELENGTH_NAME].to_numpy()

    def get_bands_responses(self, options=None):
        """
        Retrieves the bands responses.

        Parameters
        ----------
        options : S2SrfOptions
            The satellite, band names and wavelength range.
            If satellite is missing, satellite 'A' will be used.
            If band ids are missing, all band ids will be used.
            If wavelength range is missing, (360, 830) will be used.

        Returns
        -------
        output : ndarray
            A (band_names_size x wavelengths_size) array containing
            the spectral responses of the given bands.
        """
        satellite, band_ids, wavelength_range = self._parse_s2srf_options(options)

        wavelengths = self.get_wavelengths()
        mask = (wavelengths >= wavelength_range[0]) & (wavelengths <= wavelength_range[1])
        band_names = self.get_band_names(band_ids, satellite)

        return self.s2_srf_data[satellite][band_names].T.to_numpy()[:, mask]

    def _parse_s2srf_options(self, options):
        if options is None:
            satellite, band_ids, wavelength_range = None, None, None
        else:
            satellite = options.satellite
            band_ids = options.band_ids
            wavelength_range = options.wavelength_range
        if satellite is None:
            satellite = 'A'
        if band_ids is None:
            band_ids = list(range(0, len(self._BAND_NAMES)))
        if wavelength_range is None:
            wavelength_range = (360, 830)

        return satellite, band_ids, wavelength_range

    def get_band_names(self, band_ids=None, satellite='A'):
        """
        Retrieves the band names corresponding to the provided band ids.

        Parameters
        ----------
        band_ids : list of int
            The band ids - 0 to 12.
            If missing, default to all bands.
        satellite : str
            The satellite - 'A' or 'B'.
            If missing, default to 'A'.

        Returns
        -------
        output : list
            A list containing all the band names.
        """
        if band_ids is None:
            band_ids = list(range(0, len(self._BAND_NAMES)))
        return self.all_band_names[satellite][band_ids]

    def get_bands_responses_distribution(self, options=None):
        """
        Read the Sentinel-2 SRF values to a colour.MultiSpectralDistributions

        Parameters
        ----------
        options : S2SrfOptions
            The satellite, band names and wavelength range.
            If satellite is missing, satellite 'A' will be used.
            If band names are missing, all band names will be used.
            If wavelength range is missing, (360, 830) will be used.

        Returns
        -------
        output : colour.MultiSpectralDistributions
            The Sentinel-2 spectral response functions
            as a colour.MultiSpectralDistributions object.
        """
        satellite, band_ids, wavelength_range = self._parse_s2srf_options(options)

        band_names = self.get_band_names(band_ids, satellite)

        wavelengths = self.get_wavelengths()
        mask = (wavelengths >= wavelength_range[0]) & (wavelengths <= wavelength_range[1])
        wavelengths = wavelengths[mask]

        bands_srf = self.s2_srf_data[satellite][band_names].to_numpy()[mask, :]

        return MultiSpectralDistributions(dict(zip(wavelengths, bands_srf)))



"""
S2Product
================

S2Product module provides tools for working with Sentinel-2 product data.

s2_product provides the class S2Product that has methods for working with
band data - for example: read specific band, convert dn to sentinel,
get the metadata.

"""

from sentinel_toolkit.product.s2_product import S2Product
from sentinel_toolkit.product.product_metadata import S2ProductMetadata



import os
import unittest

import numpy as np
from numpy.testing import assert_array_equal

from sentinel_toolkit.product import S2ProductMetadata


class TestProductMetadata(unittest.TestCase):
    _METADATA_FILENAME = os.path.join(os.path.dirname(__file__), "test_data", "metadata.xml")

    _EXPECTED_SENSING_DATE = "2022-09-15"
    _EXPECTED_NODATA_VALUE = 0
    _EXPECTED_BOA_QUANTIFICATION_VALUE = 10000
    _EXPECTED_BAND_ID_TO_BOA_OFFSET = {
        0: -1000,
        1: -1000,
        2: -1000,
        3: -1000
    }
    _EXPECTED_BAND_ID_TO_SOLAR_IRRADIANCE = {
        0: 1874.3,
        1: 1959.75,
        2: 1824.93,
        3: 1512.79
    }

    def setUp(self):
        self.product_metadata = S2ProductMetadata(self._METADATA_FILENAME)

    def test_get_sensing_date(self):
        actual = self.product_metadata.get_sensing_date()
        self.assertEqual(self._EXPECTED_SENSING_DATE, actual)

    def test_get_nodata_value(self):
        actual = self.product_metadata.get_nodata_value()
        self.assertEqual(self._EXPECTED_NODATA_VALUE, actual)

    def test_get_boa_quantification_value(self):
        actual = self.product_metadata.get_quantification_value()
        self.assertEqual(self._EXPECTED_BOA_QUANTIFICATION_VALUE, actual)

    def test_get_band_id_to_boa_offset(self):
        actual = self.product_metadata.get_band_id_to_offset()
        self.assertEqual(self._EXPECTED_BAND_ID_TO_BOA_OFFSET, actual)

    def test_get_boa_offsets(self):
        band_ids = [1, 2, 3]
        expected = np.array([self._EXPECTED_BAND_ID_TO_BOA_OFFSET[band_id] for band_id in band_ids])
        actual = self.product_metadata.get_offsets(band_ids)
        assert_array_equal(expected, actual)

    def test_get_band_id_to_solar_irradiance(self):
        actual = self.product_metadata.get_band_id_to_solar_irradiance()
        self.assertEqual(self._EXPECTED_BAND_ID_TO_SOLAR_IRRADIANCE, actual)

    def test_get_solar_irradiances(self):
        band_ids = [1, 2, 3]
        expected = np.array([self._EXPECTED_BAND_ID_TO_SOLAR_IRRADIANCE[band_id] for band_id in band_ids])
        actual = self.product_metadata.get_solar_irradiances(band_ids)
        assert_array_equal(expected, actual)

    def test_get_normalized_solar_irradiances(self):
        band_ids = [1, 2, 3]
        expected_solar_irradiance = np.array([self._EXPECTED_BAND_ID_TO_SOLAR_IRRADIANCE[band_id]
                                              for band_id in band_ids])
        expected = expected_solar_irradiance / np.max(expected_solar_irradiance)
        actual = self.product_metadata.get_normalized_solar_irradiances(band_ids)
        assert_array_equal(expected, actual)


if __name__ == '__main__':
    unittest.main()



"""
S2Product Metadata
================

S2Product Metadata module provides tools for working with
Sentinel-2 product metadata.

s2_metadata provides the class S2ProductMetadata that has methods
for retrieving information like Sensing Date, NODATA Value,
Quantification Value, BOA Offset Values, Solar Irradiances.
"""

from sentinel_toolkit.product.product_metadata.s2_metadata import S2ProductMetadata



"""
ecostress provides a wrapper class around spectral.EcostressDatabase.
"""

import numpy as np
from colour import SpectralDistribution, SpectralShape, LinearInterpolator

from scipy.interpolate import interp1d

from sentinel_toolkit.colorimetry.sentinel_values import SpectralData


class Ecostress:
    """
    Ecostress is a wrapper around the EcostressDatabase from spectral library.
    The extra functionality that is provided is querying products that
    have some values in a given wavelength range.
    """

    def __init__(self, ecostress_db):
        self.ecostress_db = ecostress_db

    def get_spectrum_ids(self, wavelength_range=None):
        """
        Retrieves the spectrum identifiers of the ecostress examples.

        Ony examples that have some spectral data in the
        given `wavelength_range` are returned.

        Parameters
        ----------
        wavelength_range : tuple of int
            The wavelength range.

        Returns
        -------
        output : list of int
            A list of the identifiers of the found examples.
        """
        if wavelength_range is None:
            wavelength_range = (360, 830)

        sql = """
        select SpectrumID 
        from Spectra
        where MinWaveLength <= ? and MaxWaveLength >= ?
         """

        min_wavelength = wavelength_range[0] / 1000
        max_wavelength = wavelength_range[1] / 1000

        result = self.ecostress_db.query(sql, (max_wavelength, min_wavelength)).fetchall()
        return [r[0] for r in result]

    def get_spectral_distribution_colour(self, spectrum_id, wavelength_range=None):
        """
        Retrieves the SpectralDistribution.

        The SpectralDistribution with id `spectrum_id` is retrieved
        and it is trimmed to the provided `wavelength_range`.

        Parameters
        ----------
        spectrum_id : int
            The spectrum identifier.
        wavelength_range : tuple of int
            The wavelength range.

        Returns
        -------
        output : colour.SpectralDistribution
            The corresponding SpectralDistribution.
        """
        if wavelength_range is None:
            wavelength_range = (360, 830)

        signature = self.ecostress_db.get_signature(spectrum_id)

        wavelengths = np.round(np.array(signature.x), 4) * 1000
        spectral_responses = np.round(np.array(signature.y), 4) / 100

        spectral_data = dict(zip(wavelengths, spectral_responses))

        spectral_distribution = SpectralDistribution(spectral_data, name='Ecostress')
        shape = SpectralShape(wavelength_range[0], wavelength_range[1], 1)
        spectral_distribution.interpolate(shape, LinearInterpolator)

        return spectral_distribution

    def get_spectral_distribution_numpy(self, spectrum_id, wavelength_range=None):
        """
        Retrieves the SpectralDistribution.

        The SpectralDistribution with id `spectrum_id` is retrieved
        and it is trimmed to the provided `wavelength_range`.

        This method can be used for gaining better performance, avoiding
        the overhead of creating colour.SpectralDistribution.

        Parameters
        ----------
        spectrum_id : int
            The spectrum identifier.
        wavelength_range : tuple of int
            The wavelength range.

        Returns
        -------
        output : SpectralData (tuple)
            The tupled wavelengths and spectral_responses.
        """
        if wavelength_range is None:
            wavelength_range = (360, 830)

        signature = self.ecostress_db.get_signature(spectrum_id)

        wavelengths = np.trunc(np.round(np.array(signature.x), 4) * 1000).astype(int)
        spectral_responses = np.round(np.array(signature.y), 4) / 100

        interpolator = interp1d(wavelengths, spectral_responses)

        min_wavelength = max(wavelengths[0], wavelength_range[0])
        max_wavelength = min(wavelengths[-1], wavelength_range[1])

        wavelengths = np.arange(min_wavelength, max_wavelength + 1, 1)
        spectral_responses = interpolator(wavelengths)

        return SpectralData(wavelengths, spectral_responses)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sentinel_toolkit',
 'sentinel_toolkit.colorimetry',
 'sentinel_toolkit.colorimetry.illuminants',
 'sentinel_toolkit.colorimetry.tests',
 'sentinel_toolkit.converter',
 'sentinel_toolkit.converter.tests',
 'sentinel_toolkit.ecostress',
 'sentinel_toolkit.ecostress.tests',
 'sentinel_toolkit.product',
 'sentinel_toolkit.product.product_metadata',
 'sentinel_toolkit.product.product_metadata.tests',
 'sentinel_toolkit.product.tests',
 'sentinel_toolkit.srf',
 'sentinel_toolkit.srf.tests']

package_data = \
{'': ['*'],
 'sentinel_toolkit.product.product_metadata.tests': ['test_data/*'],
 'sentinel_toolkit.product.tests': ['test_data/S2B_MSIL2A_20220915T092029_N0400_R093_T34TGN_20220915T112452.SAFE/*',
                                    'test_data/S2B_MSIL2A_20220915T092029_N0400_R093_T34TGN_20220915T112452.SAFE/GRANULE/L2A_T34TGN_A028862_20220915T092030/IMG_DATA/R10m/*',
                                    'test_data/S2B_MSIL2A_20220915T092029_N0400_R093_T34TGN_20220915T112452.SAFE/GRANULE/L2A_T34TGN_A028862_20220915T092030/IMG_DATA/R20m/*'],
 'sentinel_toolkit.srf.tests': ['test_data/*']}

install_requires = \
['colour-science>=0.4.1,<0.5.0',
 'elementpath>=3.0.2,<4.0.0',
 'openpyxl>=3.0.10,<4.0.0',
 'pandas>=1.4.2,<2.0.0',
 'rasterio>=1.2,<2.0',
 'spectral>=0.22.4,<0.24.0']

setup_kwargs = {
    'name': 'sentinel-toolkit',
    'version': '7.0.3',
    'description': 'Various sentinel tools',
    'long_description': '# Sentinel-Toolkit\n\n# Description\n\nThis repository provides various utility tools for working with Sentinel data like:\n\n1. Converting raw band dn values to sentinel responses.\n2. Wrapper classes around a Sentinel-2 Product and Metadata.\n3. Generating SQLite database for a given Ecostress Spectral Library.\n4. Wrapper class around spectral.EcostressDatabase for querying with specific filters.\n5. Wrapper class around Sentinel-2 Spectral Response Functions Excel file.\n6. Converting a spectral distribution to Sentinel Responses.\n7. Converting Ecostress Spectral Library to Sentinel Responses CSV file.\n\n# Installation\n\nSentinel-Toolkit and its primary dependencies can be easily installed from the Python Package Index by issuing this\ncommand in a shell:\n\n```shell\n$ pip install --user sentinel-toolkit\n```\n\n# Examples\n\n## Converting Sentinel-2 DN values to Sentinel Responses\n\nThe new conversion formula for products after 25 January 2022 is integrated.\n\nFor more info\ncheck https://forum.step.esa.int/t/changes-in-band-data-after-25-jan-2022-baseline-04-00-harmonizevalues-sentinel-2-l2a-snappy/36270\n\n```python\nfrom sentinel_toolkit.colorimetry import dn_to_sentinel\n\n# The raw band(s) data\n# For example: with 3 bands, 10980 x 10980 x 3 ndarray can be passed\nbands_dn = None\n\n# These values can be retrieved from product.S2ProductMetadata\n# and this algorithm can be run easier from an S2Product object.\n# See section "Working with Sentinel-2 Product".\nnodata_value = 0\nbands_offsets = [-1000, -1000, -1000]\nquantification_value = 10000\nnormalized_solar_irradiances = [1, 0.9312, 0.7719]\n\nsentinel_responses = dn_to_sentinel(bands_dn,\n                                    nodata_value,\n                                    bands_offsets,\n                                    quantification_value,\n                                    normalized_solar_irradiances)\n```\n\n## Working with Sentinel-2 Product and Metadata\n\n### Working with Sentinel-2 Product Metadata\n\n```python\nfrom sentinel_toolkit.product import S2ProductMetadata\n\nproduct_metadata = S2ProductMetadata("<path-to-metadata-filename>")\n\n# With S2ProductMetadata you can retrieve:\n# sensing data, nodata value, quantification value, band offsets\n# and solar irradiances. Currently, these methods read directly\n# from the xml and there is no caching.\nproduct_metadata.get_sensing_date()\nproduct_metadata.get_nodata_value()\nproduct_metadata.get_quantification_value()\nproduct_metadata.get_band_id_to_offset()\nproduct_metadata.get_band_id_to_solar_irradiance()\n\nband_ids = [1, 2, 3]\nproduct_metadata.get_offsets(band_ids)\nproduct_metadata.get_solar_irradiances(band_ids)\nproduct_metadata.get_normalized_solar_irradiances(band_ids)\n```\n\n### Working with Sentinel-2 Product\n\n```python\nfrom sentinel_toolkit.product import S2Product\n\nproduct = S2Product("<path-to-product-directory>")\n\nproduct.get_directory_name()\nproduct.get_metadata()\n\n# If multiple sources are available for a given band\n# for example B02_10m.jp2, B02_20m.jp2, B02_60m.jp2,\n# the file with the best resolution will be selected.\n# In the example case, this is B02_10m.jp2.\n# (Currently this is implemented based on the band suffix).\nproduct.get_dn_source(band_id=1)\n\nband_ids = [1, 2, 3]\nproduct.get_band_id_to_dn_source(band_ids)\n\n# Converts the given bands to sentinel responses by\n# using colorimetry.dn_to_sentinel() passing the required\n# metadata arguments, so you don\'t have to do it yourself.\nout_shape = (10980, 10980)\nprofile, sentinel_responses = product.dn_to_sentinel(band_ids, out_shape)\n```\n\n## Loading and working with Ecostress Spectral Library\n\n### Generating SQLite database\n\nGenerate the SQLite database given the Ecostress spectral library directory:\n\n```python\nfrom sentinel_toolkit.ecostress import generate_ecostress_db\n\ngenerate_ecostress_db("ecospeclib-all", "ecostress.db")\n```\n\nFor convenience, there is a main method in ecostress_db_generator.py that can be called from shell like so:\n\n```shell\n$ python ecostress_db_generator.py -d /ecospeclib-all -o ecostress.db\n```\n\n### Working with the generated SQLite database\n\n```python\nfrom spectral import EcostressDatabase\nfrom sentinel_toolkit.ecostress import Ecostress\n\necostress_db = EcostressDatabase("ecostress.db")\necostress = Ecostress(ecostress_db)\n\n# Get all the spectrum ids that contain some values in the given range (420, 830).\nwavelength_range = (420, 830)\nspectrum_ids = ecostress.get_spectrum_ids(wavelength_range)\n\n# Iterate over the found spectrum_ids and get colour.SpectralDistribution objects.\nspectral_distributions_colour = []\nfor spectrum_id in spectrum_ids:\n    spectral_distribution = ecostress.get_spectral_distribution_colour(spectrum_id)\n    spectral_distributions_colour.append(spectral_distribution)\n\n# Iterate over the found spectrum_ids and get numpy arrays.\n# This can be used for gaining better performance\nspectral_distributions_numpy = []\nfor spectrum_id in spectrum_ids:\n    spectral_distribution = ecostress.get_spectral_distribution_numpy(spectrum_id)\n    spectral_distributions_numpy.append(spectral_distribution)\n```\n\n## Reading Sentinel-2 Spectral Response Functions\n\nGiven an Excel file containing the Sentinel-2 Spectral Response Functions, retrieve the wavelengths, band names and\nbands_responses as colour.MultiSpectralDistributions and 2D ndarray:\n\n```python\nfrom sentinel_toolkit.srf import S2Srf, S2SrfOptions\n\ns2a_srf = S2Srf("srf.xlsx")\n\n# Retrieve the wavelengths of Sentinel-2A as ndarray.\nwavelengths = s2a_srf.get_wavelengths()\n\n# Retrieve all the band names of Sentinel-2A bands 1, 2 and 3 as ndarray.\nband_names = s2a_srf.get_band_names(band_ids=[1, 2, 3])\n\n# Retrieve B2, B3, B4 of Sentinel-2A satellite in wavelength range (360, 830)\n# as colour.MultiSpectralDistributions.\nsatellite = \'A\'\nband_ids_option = [1, 2, 3]\nwavelength_range = (360, 830)\ns2_srf_options = S2SrfOptions(satellite, band_ids_option, wavelength_range)\nbands_responses_distribution = s2a_srf.get_bands_responses_distribution(s2_srf_options)\n\n# Retrieve all bands responses of Sentinel-2B in wavelength range (360, 830) as 2D ndarray.\nsatellite = \'B\'\nwavelength_range = (360, 830)\ns2_srf_options = S2SrfOptions(satellite=satellite, wavelength_range=wavelength_range)\nbands_responses = s2a_srf.get_bands_responses(s2_srf_options)\n```\n\n## Converting SpectralDistribution to Sentinel-2 Responses\n\nConvert a spectral distribution to Sentinel-2 Responses:\n\n```python\nfrom spectral import EcostressDatabase\nfrom sentinel_toolkit.ecostress import Ecostress\nfrom sentinel_toolkit.srf import S2Srf, S2SrfOptions\nfrom sentinel_toolkit.colorimetry import sd_to_sentinel_numpy, sd_to_sentinel_direct_numpy\nfrom sentinel_toolkit.colorimetry.illuminants import D65_360_830_1NM_VALUES\n\necostress_db = EcostressDatabase("ecostress.db")\necostress = Ecostress(ecostress_db)\ns2a_srf = S2Srf("srf.xlsx")\n\nwavelength_range = (360, 830)\n\nspectrum_id = 1\n# Use the numpy version for better performance\nspectral_data = ecostress.get_spectral_distribution_numpy(spectrum_id, wavelength_range)\n\nspectral_data_min_wavelength = spectral_data.wavelengths[0]\nspectral_data_max_wavelength = spectral_data.wavelengths[-1]\n\nwr_start = max(wavelength_range[0], spectral_data_min_wavelength)\nwr_end = min(wavelength_range[1], spectral_data_max_wavelength)\n\n# Reshape the illuminant to the spectral distribution shape\nilluminant = D65_360_830_1NM_VALUES[wr_start - 360: wr_end - 359]\n\n# Get the sentinel responses for spectrum with id 1 for all bands\n# from satellite \'A\' in wavelength_range (360, 830)\ns2_srf_options = S2SrfOptions(satellite=\'A\', wavelength_range=(wr_start, wr_end))\nsentinel_responses = sd_to_sentinel_numpy(spectral_data,\n                                          s2a_srf,\n                                          s2_srf_options,\n                                          illuminant)\n\n# Another way of doing this would be:\ns2_srf_options = S2SrfOptions(satellite=\'A\', wavelength_range=(wr_start, wr_end))\nbands_responses = s2a_srf.get_bands_responses(s2_srf_options)\nsentinel_responses = sd_to_sentinel_direct_numpy(spectral_data, bands_responses, illuminant)\n```\n\n## Converting full Ecostress Spectral Library to Sentinel-2 Responses CSV file\n\nGenerate a CSV file containing the Sentinel-2 responses for all materials from the Ecostress library:\n\n```python\nfrom spectral import EcostressDatabase\nfrom sentinel_toolkit.ecostress import Ecostress\nfrom sentinel_toolkit.srf import S2Srf\nfrom sentinel_toolkit.converter import EcostressToSentinelConverter\n\necostress_db = EcostressDatabase("ecostress.db")\necostress = Ecostress(ecostress_db)\n\ns2a_srf = S2Srf("srf.xlsx")\n\nconverter = EcostressToSentinelConverter(ecostress, s2a_srf)\nconverter.convert_ecostress_to_sentinel_csv()\n```\n\nFor convenience, there is a main method in converter.py that can be called from shell like so:\n\n```shell\n$ python converter.py -e ecostress.db -s2 S2-SRF_COPE-GSEG-EOPG-TN-15-0007_3.0.xlsx -s A -b 1 2 3 -ws 360 -we 830\n```\n',
    'author': 'Georgi Genchev',
    'author_email': 'gdgenchev97@gmail.com',
    'maintainer': 'Georgi Genchev',
    'maintainer_email': 'gdgenchev97@gmail.com',
    'url': 'https://github.com/sentinel-toolkit/sentinel-toolkit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import unittest
from unittest.mock import patch

import numpy as np
from spectral import EcostressDatabase
from spectral.database.aster import Signature

from sentinel_toolkit.ecostress import Ecostress

from numpy.testing import assert_array_equal
from numpy.testing import assert_array_almost_equal


class TestEcostress(unittest.TestCase):
    _DB_FILENAME = "ecostress.db"
    _ECOSTRESS_DATA_DIRECTORY = "ecospeclib-all"

    _QUERY = """
        select SpectrumID 
        from Spectra
        where MinWaveLength <= ? and MaxWaveLength >= ?
         """

    _SPECTRUM_ID = 0

    _SIGNATURE_LEN_6 = Signature()
    _SIGNATURE_LEN_6.x = np.array([0.438, 0.439, 0.440, 0.441, 0.442, 0.443])
    _SIGNATURE_LEN_6.y = np.array([10.4000, 10.4200, 10.4300, 10.4400, 10.4500, 10.4600])

    _SIGNATURE_LEN_7 = Signature()
    _SIGNATURE_LEN_7.x = np.array([0.438, 0.439, 0.440, 0.441, 0.442, 0.443, 0.444])
    _SIGNATURE_LEN_7.y = np.array([10.4000, 10.4200, 10.4300, 10.4400, 10.4500, 10.4600, 10.4800])

    @patch.object(EcostressDatabase, 'query')
    def test_get_spectrum_ids(self, mock_ecostress_database_query):
        ecostress = Ecostress(EcostressDatabase())
        ecostress.get_spectrum_ids()
        mock_ecostress_database_query.assert_called_once_with(self._QUERY, (0.83, 0.36))

    @patch.object(EcostressDatabase, 'query')
    def test_get_spectrum_ids_with_wavelength_range(self, mock_ecostress_database_query):
        ecostress = Ecostress(EcostressDatabase())
        ecostress.get_spectrum_ids(wavelength_range=(420, 440))
        mock_ecostress_database_query.assert_called_once_with(self._QUERY, (0.44, 0.42))

    @patch.object(EcostressDatabase, 'get_signature')
    def test_get_spectral_distribution_colour(self, mock_ecostress_db):
        mock_ecostress_db.get_signature.return_value = self._SIGNATURE_LEN_6

        ecostress = Ecostress(mock_ecostress_db)
        sd = ecostress.get_spectral_distribution_colour(self._SPECTRUM_ID)

        assert_array_equal(self._SIGNATURE_LEN_6.x * 1000, sd.wavelengths)
        assert_array_almost_equal(self._SIGNATURE_LEN_6.y / 100, sd.values)

    @patch.object(EcostressDatabase, 'get_signature')
    def test_get_spectral_distribution_colour_with_wavelength_range(self, mock_ecostress_db):
        mock_ecostress_db.get_signature.return_value = self._SIGNATURE_LEN_7

        ecostress = Ecostress(mock_ecostress_db)
        sd = ecostress.get_spectral_distribution_colour(self._SPECTRUM_ID, wavelength_range=(438, 443))

        assert_array_equal(self._SIGNATURE_LEN_7.x[0:-1] * 1000, sd.wavelengths)
        assert_array_almost_equal(self._SIGNATURE_LEN_7.y[0:-1] / 100, sd.values)

    @patch.object(EcostressDatabase, 'get_signature')
    def test_get_spectral_distribution_numpy(self, mock_ecostress_db):
        mock_ecostress_db.get_signature.return_value = self._SIGNATURE_LEN_6

        ecostress = Ecostress(mock_ecostress_db)
        sd = ecostress.get_spectral_distribution_colour(self._SPECTRUM_ID)

        assert_array_equal(self._SIGNATURE_LEN_6.x * 1000, sd.wavelengths)
        assert_array_almost_equal(self._SIGNATURE_LEN_6.y / 100, sd.values)

    @patch.object(EcostressDatabase, 'get_signature')
    def test_get_spectral_distribution_numpy_with_wavelength_range(self, mock_ecostress_db):
        mock_ecostress_db.get_signature.return_value = self._SIGNATURE_LEN_7

        ecostress = Ecostress(mock_ecostress_db)
        wavelengths, spectral_responses = ecostress.get_spectral_distribution_numpy(self._SPECTRUM_ID,
                                                                                    wavelength_range=(438, 443))

        assert_array_equal(self._SIGNATURE_LEN_7.x[0:-1] * 1000, wavelengths)
        assert_array_almost_equal(self._SIGNATURE_LEN_7.y[0:-1] / 100, spectral_responses)


if __name__ == '__main__':
    unittest.main()
